#
#  Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an
# express license agreement from NVIDIA CORPORATION is strictly
# prohibited.
#

from training_telemetry.application_type import ApplicationType
from training_telemetry.config import (
    BackendConfig,
    BackendType,
    FileBackendConfig,
    LoggerBackendConfig,
    NVTXBackendConfig,
    OTLPLogsBackendConfig,
    OTLPLogsFormat,
    OTLPTracesBackendConfig,
    TelemetryConfig,
)
from training_telemetry.config_loader import load_config, load_config_from_dict, load_config_from_env
from training_telemetry.context import (
    checkpoint_save,
    get_recorder,
    running,
    training,
)
from training_telemetry.duration import Duration
from training_telemetry.events import Event, EventName, ExceptionEvent
from training_telemetry.metrics import (
    ApplicationMetrics,
    CheckpointMetrics,
    CheckPointType,
    EventMetric,
    EventMetrics,
    IterationMetrics,
)
from training_telemetry.provider import Provider
from training_telemetry.recorder import Recorder
from training_telemetry.spans import Span, SpanColor, SpanName

# Deprecated, use get_rank_index and get_rank_count
from training_telemetry.torch.utils import get_rank, get_world_size, is_rank0
from training_telemetry.utils import (
    get_current_time,
    get_elapsed_time,
    get_logger,
    get_rank_count,
    get_rank_index,
    get_timestamp_in_local_timezone,
)
from training_telemetry.verbosity import Verbosity
from training_telemetry.version import __version__



#
#  Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an
# express license agreement from NVIDIA CORPORATION is strictly
# prohibited.
#
import os
import tempfile
from pathlib import Path
from typing import Any

import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader, TensorDataset
from tzlocal import get_localzone

from training_telemetry.config import BackendType
from training_telemetry.config_loader import load_config
from training_telemetry.context import checkpoint_save, get_recorder, running, timed_span, training
from training_telemetry.events import Event, EventName
from training_telemetry.metrics import (
    ApplicationMetrics,
    CheckpointMetrics,
    CheckPointType,
    DeviceMetrics,
    IterationMetrics,
    NVLinkMetrics,
)
from training_telemetry.provider import Provider
from training_telemetry.spans import SpanColor, SpanName
from training_telemetry.utils import get_rank_count, get_rank_index
from training_telemetry.verbosity import Verbosity

# Generate some random data for this example
torch.manual_seed(42)
# Generate random input features and binary labels
X = torch.randn(1000, 10, dtype=torch.float32)
y = (X.sum(dim=1) > 0).float()
dataset = TensorDataset(X, y)
dataloader = DataLoader(dataset, batch_size=32, shuffle=True, num_workers=0)
num_epochs = 100

# Initialize the telemetry provider with a default configuration
config = load_config(
    config_file=Path(__file__).parent / "example_config.yaml",
    defaults={"application": {"job_name": "torch_example", "job_id": "1234567890", "environment": "test"}},
    override_from_env=False,
)
Provider.set_provider(config)


# Define a simple model
class SimpleModel(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.layers = nn.Sequential(nn.Linear(10, 64), nn.ReLU(), nn.Linear(64, 1), nn.Sigmoid())

    def forward(self, x: torch.Tensor) -> Any:
        return self.layers(x)


def get_application_metrics() -> ApplicationMetrics:
    return ApplicationMetrics.create(
        rank=get_rank_index(),
        world_size=get_rank_count(),
        node_name="localhost",
        timezone=str(get_localzone()),
        total_iterations=num_epochs * len(dataloader),
        checkpoint_enabled=True,
        checkpoint_strategy="sync",
    )


@running(metrics=get_application_metrics())
def main() -> None:
    # Initialize model, loss function and optimizer
    model = SimpleModel()
    criterion = nn.BCELoss()
    optimizer = optim.Adam(model.parameters(), lr=0.001, weight_decay=0.0001)

    events = []
    with training() as training_span:
        current_iteration = 0
        accuracy = torch.tensor(float("nan"))
        loss = torch.tensor(float("nan"))

        for epoch in range(num_epochs):
            for batch_idx, (inputs, targets) in enumerate(dataloader):
                with timed_span(SpanName.ITERATION, color=SpanColor.RED, verbosity=Verbosity.PROFILING):
                    # Forward pass
                    with timed_span(SpanName.MODEL_FORWARD, color=SpanColor.RED, verbosity=Verbosity.PROFILING):
                        outputs = model(inputs)
                        loss = criterion(outputs.squeeze(), targets)

                    # Backward pass and optimize
                    with timed_span(SpanName.ZERO_GRAD, color=SpanColor.GREEN, verbosity=Verbosity.PROFILING):
                        optimizer.zero_grad()
                    with timed_span(SpanName.MODEL_BACKWARD, color=SpanColor.BLUE, verbosity=Verbosity.PROFILING):
                        loss.backward()
                    with timed_span(SpanName.OPTIMIZER_UPDATE, color=SpanColor.YELLOW, verbosity=Verbosity.PROFILING):
                        optimizer.step()

                    # Calculate accuracy
                    predictions = (outputs.squeeze() > 0.5).float()
                    accuracy = (predictions == targets).float().mean()

                    current_iteration += 1

                # Log iteration metrics to TRACING backends every iteration
                metrics = IterationMetrics.create(
                    current_iteration=current_iteration,
                    num_iterations=1,
                    average_iteration_time=0.5,  # dummy value, in real life the elapsed time for the iteration
                    average_forward_time=0.1,  # dummy value, in real life the elapsed time for the forward pass
                    average_backward_time=0.2,  # dummy value, in real life the elapsed time for the backward pass
                    average_dataloader_time=0.3,  # dummy value, in real life the elapsed time for the dataloader
                    tflops=123.4,  # dummy value, in real life the estimated tflops for the iteration
                    loss=loss.item(),
                )
                events.append(Event.create(EventName.TRAINING_ITERATIONS, metrics=metrics))
                get_recorder().event(
                    Event.create(EventName.TRAINING_ITERATIONS, metrics=metrics),
                    training_span,
                    verbosity=Verbosity.TRACING,
                )

                # Every 50 iterations, log the averages to the LOGGER backend only, and reset the events. Also log device metrics to all backends.
                if current_iteration % 50 == 0:
                    metrics = IterationMetrics.create(
                        current_iteration=current_iteration,
                        num_iterations=sum([e.metrics["num_iterations"].value for e in events]),
                        average_iteration_time=sum([e.metrics["average_iteration_time"].value for e in events])
                        / len(events),
                        average_forward_time=sum([e.metrics["average_forward_time"].value for e in events])
                        / len(events),
                        average_backward_time=sum([e.metrics["average_backward_time"].value for e in events])
                        / len(events),
                        average_dataloader_time=sum([e.metrics["average_dataloader_time"].value for e in events])
                        / len(events),
                        tflops=sum([e.metrics["tflops"].value for e in events]) / len(events),
                        loss=loss.item(),
                    )
                    get_recorder().event(
                        Event.create(EventName.TRAINING_ITERATIONS, metrics=metrics),
                        training_span,
                        verbosity=Verbosity.INFO,
                        backend_types={BackendType.LOGGER},
                    )
                    events = []

                    device_metrics = DeviceMetrics.create(
                        current_iteration=current_iteration,
                        gpu_utilization=0.5,  # dummy value
                        memory_usage=0.5,  # dummy value
                        power_draw=0.5,  # dummy value
                        temperature=0.5,  # dummy value
                        clock_rate=0.5,  # dummy value
                        nvlink_metrics=NVLinkMetrics.create(
                            tx_data_payload=1000,  # dummy value
                            rx_data_payload=1000,
                            crc_errors=0,  # dummy value
                            replay_errors=0,  # dummy value
                            recovery_errors=0,  # dummy value
                        ),
                    )
                    get_recorder().event(
                        Event.create(EventName.DEVICE_PROPERTIES, metrics=device_metrics),
                        training_span,
                        verbosity=Verbosity.INFO,
                    )

            # Save checkpoint every 5 epochs
            if epoch % 5 == 0:
                print(
                    f"Epoch [{epoch+1}/{num_epochs}], "
                    f"Batch [{batch_idx+1}/{len(dataloader)}], "
                    f"Loss: {loss.item():.4f}, "
                    f"Accuracy: {accuracy.item():.4f}"
                )
                checkpoint = {
                    "model_state_dict": model.state_dict(),
                    "optimizer_state_dict": optimizer.state_dict(),
                    "epoch": epoch,
                    "iteration": current_iteration,
                    "loss": loss.item(),
                    "accuracy": accuracy.item(),
                }
                with checkpoint_save() as checkpoint_save_span:
                    with tempfile.TemporaryDirectory() as temp_dir:
                        checkpoint_file_name = os.path.join(temp_dir, f"checkpoint_iter_{current_iteration}.pt")
                        torch.save(checkpoint, checkpoint_file_name)
                        checkpoint_save_span.add_metrics(
                            CheckpointMetrics.create(
                                checkpoint_type=CheckPointType.LOCAL,
                                current_iteration=current_iteration,
                                num_iterations=len(dataloader),
                                checkpoint_directory=temp_dir,
                            )
                        )


if __name__ == "__main__":
    main()



#
#  Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an
# express license agreement from NVIDIA CORPORATION is strictly
# prohibited.
#
from enum import Enum


class ApplicationType(str, Enum):
    """
    The type of training application, such as pytorch, jax, etc.
    Currently, only pytorch and jax are supported
    """

    # This is supported and the default.
    PYTORCH = "pytorch"
    # Support for this will be added by https://jirasw.nvidia.com/browse/HEIMDALL-1310
    JAX = "jax"
    # This is not supported yet.
    TENSORFLOW = "tensorflow"



#
#  Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an
# express license agreement from NVIDIA CORPORATION is strictly
# prohibited.
#
import os
import tempfile
from pathlib import Path
from typing import Any

import lightning as L
import torch
import torch.nn as nn
from lightning.pytorch.callbacks import ModelCheckpoint
from torch.utils.data import DataLoader, TensorDataset

from training_telemetry.config_loader import load_config
from training_telemetry.torch.lightning.callback import TelemetryCallback


# Define a Lightning model
class SimpleModel(L.LightningModule):
    def __init__(self, learning_rate: float = 0.001, weight_decay: float = 0.0001) -> None:
        super().__init__()
        self.save_hyperparameters()
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay

        # Model architecture
        self.layers = nn.Sequential(nn.Linear(10, 64), nn.ReLU(), nn.Linear(64, 1), nn.Sigmoid())

        # Loss function
        self.criterion = nn.BCELoss()

    def forward(self, x: torch.Tensor) -> Any:
        return self.layers(x)

    def training_step(self, batch: tuple[torch.Tensor, torch.Tensor], batch_idx: int) -> Any:
        inputs, targets = batch

        # Forward pass
        outputs = self(inputs)
        loss = self.criterion(outputs.squeeze(), targets)

        # Calculate accuracy
        predictions = (outputs.squeeze() > 0.5).float()
        accuracy = (predictions == targets).float().mean()

        # Log metrics
        self.log("train_loss", loss, prog_bar=True)
        self.log("train_accuracy", accuracy, prog_bar=True)

        return loss

    def configure_optimizers(self) -> Any:
        optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate, weight_decay=self.weight_decay)
        return optimizer


def main() -> None:
    # Set random seed for reproducibility
    L.seed_everything(42)

    # Generate some random data for this example
    X = torch.randn(1000, 10, dtype=torch.float32)
    y = (X.sum(dim=1) > 0).float()
    dataset = TensorDataset(X, y)

    log_interval = 100  # log every 100 steps
    batch_size = 32

    # Create DataLoader
    train_dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=True, num_workers=0)

    # Initialize model
    model = SimpleModel(learning_rate=0.001, weight_decay=0.0001)

    # Initialize telemetry callback
    config = load_config(
        config_file=Path(__file__).parent / "../example_config.yaml",
        defaults={
            "application": {
                "job_name": "lightning_example",
                "job_id": "1234567890",
                "environment": "test",
                "log_interval": log_interval,
            }
        },
        override_from_env=False,
    )
    telemetry_callback = TelemetryCallback(config=config)
    telemetry_callback.on_app_start()

    temp_dir = tempfile.TemporaryDirectory()
    checkpoint_dir = os.path.join(temp_dir.name, "checkpoints")
    os.makedirs(checkpoint_dir, exist_ok=True)

    try:
        # Configure checkpoint callback
        checkpoint_callback = ModelCheckpoint(
            dirpath=checkpoint_dir,
            filename="lightning-{epoch:02d}-{train_loss:.2f}",
            every_n_epochs=20,
            save_top_k=3,
            monitor="train_loss",
            mode="min",
        )

        # Configure trainer with callbacks
        trainer = L.Trainer(
            max_epochs=100,
            enable_checkpointing=True,
            enable_progress_bar=False,
            enable_model_summary=True,
            log_every_n_steps=log_interval,
            callbacks=[telemetry_callback, checkpoint_callback],
            logger=False,
        )

        # Train the model
        trainer.fit(model, train_dataloader)
    except Exception as e:
        telemetry_callback.on_exception(trainer, model, e)
        raise e
    finally:
        telemetry_callback.on_app_end()


if __name__ == "__main__":
    main()



#
#  Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an
# express license agreement from NVIDIA CORPORATION is strictly
# prohibited.
#

import functools
import logging
import socket
import traceback
from collections import defaultdict
from dataclasses import dataclass
from typing import Any, List, Optional, Type

import torch
from lightning.fabric.utilities.types import _PATH
from lightning.pytorch import LightningModule, Trainer
from lightning.pytorch.callbacks import Callback
from lightning.pytorch.plugins.io import AsyncCheckpointIO
from lightning.pytorch.utilities.data import extract_batch_size
from lightning.pytorch.utilities.types import STEP_OUTPUT
from tzlocal import get_localzone

from training_telemetry.config import TelemetryConfig
from training_telemetry.events import Event, EventName
from training_telemetry.metrics import ApplicationMetrics, CheckpointMetrics, IterationMetrics
from training_telemetry.provider import Provider
from training_telemetry.recorder import Recorder
from training_telemetry.spans import Span, SpanName
from training_telemetry.torch.utils import end_monitoring_flops, get_rank, get_world_size, start_monitoring_flops
from training_telemetry.utils import get_current_time, get_logger
from training_telemetry.verbosity import Verbosity

# This should be calculated as soon as this file is imported, giving us a better approximation of the application start time.
# An alternative would be to make the caller pass it to on_app_start() but the signature of on_app_start() in the Nemo BaseCallback is generic.
_START_TIME = get_current_time()
_logger = get_logger(__name__)


@dataclass
class SpanSummary:
    """
    A class for tracking the number of times a span has been stopped and the total elapsed time for that span.
    """

    num: int = 0
    elapsed: float = 0

    def add(self, elapsed: float) -> None:
        self.num += 1
        self.elapsed += elapsed

    def reset(self) -> None:
        self.num = 0
        self.elapsed = 0


class IterationState:
    """Contains state that is tracked for each iteration"""

    def __init__(self) -> None:
        self.num_samples: int = 0
        self.num_flops: float = 0
        self.is_async_checkpoint: bool = False
        self.span_summaries: dict[SpanName, SpanSummary] = defaultdict(SpanSummary)
        self.last_logged_step: int = -1

    def reset(self) -> None:
        self.num_samples = 0
        self.num_flops = 0
        for span_name in self.span_summaries.keys():
            self.span_summaries[span_name].reset()


class TelemetryCallback(Callback):
    """
    Records training telemetry metrics during the training loop by using the PyTorch Lightning's callback system.
    Also records application metrics before and after the training loop, but only if running with Nemo version >= 25.09 or
    if the user application calls these additional callbacks manually. Examples of non PTL functions are:
    - on_app_start() and on_app_end()
    - on_dataloader_init_start() and on_dataloader_init_end()
    - on_optimizer_init_start() and on_optimizer_init_end()
    - on_load_checkpoint_start() and on_load_checkpoint_end()

    Checkpoint save is intercepted by wrapping the trainer.save_checkpoint method to add telemetry callbacks before and after
    the checkpoint save operation.

    Args:
        config: The telemetry configuration. This determines how the telemetry is reported, e.g. using python logging or a
         JSON file, or OpenTelemetry. It also determines the application properties, e.g. the job name, job ID, and environment
         where the application is running.
        batch_flops: The floating point operations per batch of the training data for the forward and backward passes.
          If they are known from first principles, they can be provided using this parameter, otherwise the callback
          will try to calculate them with pytorch utitlities but this is not guaranteed to work every time. If the batch flops
          chnage overtime, then this parameters should not be passed, or the callback property batch_flops should be updated manually.
        async_io_checkpoint_classes: The classes of checkpoint IO that are asynchronous.
         Pass a list of classes if using asynchronous checkpointing, so that the correct
         checkpoint strategy can be reported.

    See Also:
        https://lightning.ai/docs/pytorch/stable/common/lightning_module.html#hooks
    """

    def __init__(
        self,
        config: TelemetryConfig,
        batch_flops: float = 0,
        async_io_checkpoint_classes: List[Type[Any]] | None = None,
    ):
        super().__init__()
        self.config = config
        self.log_interval = config.application.log_interval
        self.batch_flops = batch_flops
        self.batch_flops_monitor: Optional[Any] = None
        self.async_io_checkpoint_classes = async_io_checkpoint_classes or []
        self.state = IterationState()
        self._recorder: Optional[Recorder] = (
            None  # Nemo needs this to be pickleable and it isn't, so need to postpone creating it
        )
        self._spans: dict[SpanName, Span] = {}

    def _start_span(
        self,
        span_name: Any,
        metrics: Any = None,
        verbosity: Any = None,
        color: Any = None,
        start_time: Optional[float] = None,
    ) -> Span:
        """Start a span with the given event name and optional metrics, color, and verbosity"""
        if span_name in self._spans:
            _logger.warning(f"Span {span_name} already started, stopping it but this is unexpected")
            self.recorder.stop(self._spans[span_name])
            del self._spans[span_name]
        if verbosity is None:
            verbosity = Verbosity.INFO
        if start_time is None:
            start_time = get_current_time()
        span = self.recorder.start(
            name=span_name,
            color=color,
            start_time=start_time,
            verbosity=verbosity,
            metrics=metrics,
        )
        self._spans[span_name] = span
        return span

    def _stop_span(self, span_name: Any, metrics: Any = None) -> float:
        """Stop a span and return the elapsed time"""
        span = self._spans.get(span_name)
        if span is None:
            _logger.warning(f"Span {span_name} was not started, this is unexpected")
            return 0.0
        if metrics is not None:
            span.add_metrics(metrics)

        self.recorder.stop(span)
        del self._spans[span_name]
        return span.duration.elapsed

    def _add_checkpoint_callbacks(self, trainer: Trainer) -> None:
        """Wrap the trainer.save_checkpoint method to add telemetry callbacks."""
        method = trainer.save_checkpoint

        @functools.wraps(method)
        def wrapper(filepath: _PATH, *args: Any, **kwargs: Any) -> None:
            self.on_start_save_checkpoint(trainer, filepath)
            exc = None
            try:
                method(filepath, *args, **kwargs)
            except Exception as e:
                exc = e
                raise e
            finally:
                self.on_end_save_checkpoint(trainer, exc)

        wrapper.orig_method = method  # type: ignore[attr-defined]
        trainer.save_checkpoint = wrapper  # type: ignore[method-assign]

    def _start_monitoring_flops(self) -> None:
        """
        Start monitoring FLOPS using pytorch utilities.
        This is called at the start of the first batch, unless the FLOPS have been set manually by the user.
        """
        assert self.batch_flops_monitor is None
        self.batch_flops_monitor = start_monitoring_flops()

    def _end_monitoring_flops(self) -> None:
        """
        End monitoring FLOPS using pytorch utilities. This is called at the beginning of the optimizer step of the first training batch,
        and at the end of the first training batch, in case the optimizer step callback is not called.
        """
        assert self.batch_flops_monitor is not None
        self.batch_flops = end_monitoring_flops(self.batch_flops_monitor)
        self.batch_flops_monitor = None
        if self.batch_flops == 0:
            self.batch_flops = -1  # This ensures we don't try again on the second batch if we failed to get the FLOPs

    @property
    def recorder(self) -> Recorder:
        if self._recorder is None:
            self._recorder = Provider.set_provider(self.config).recorder
        return self._recorder

    def setup(self, trainer: Trainer, pl_module: LightningModule, stage: str) -> None:
        if stage != "fit":
            return

        if trainer.save_checkpoint is not None:
            self._add_checkpoint_callbacks(trainer)

        if isinstance(trainer.strategy.checkpoint_io, AsyncCheckpointIO):
            self.state.is_async_checkpoint = True
        elif any(isinstance(trainer.strategy.checkpoint_io, klass) for klass in self.async_io_checkpoint_classes):
            self.state.is_async_checkpoint = True
        else:
            self.state.is_async_checkpoint = False

        self.on_app_start()
        self.record_app_metrics(trainer)

    def teardown(self, trainer: Trainer, pl_module: LightningModule, stage: str) -> None:
        if stage != "fit":
            return

        self.on_app_end()

        # Restore the save checkpoint method
        if trainer.save_checkpoint is not None and hasattr(trainer.save_checkpoint, "orig_method"):
            trainer.save_checkpoint = trainer.save_checkpoint.orig_method  # type: ignore[method-assign]

    def record_app_metrics(self, trainer: Trainer) -> None:
        span = self._spans.get(SpanName.MAIN_FUNCTION)
        if span is None:
            _logger.warning("Main function span not found, this is unexpected")
            return

        event_name = EventName.SPAN_ATTRIBUTES
        checkpoint_enabled = bool(trainer.checkpoint_callbacks)
        metrics = ApplicationMetrics.create(
            rank=get_rank(),
            world_size=get_world_size(),
            node_name=socket.gethostname(),
            timezone=str(get_localzone()),
            total_iterations=int(trainer.estimated_stepping_batches),
            checkpoint_enabled=checkpoint_enabled,
        )
        self.recorder.event(Event.create(event_name, metrics), span)

    # App lifecycle (from Nemo BaseCallback version >= 25.09), or otherwise called from setup() and teardown()
    def on_app_start(self, *args, **kwargs) -> None:  # type: ignore [no-untyped-def]
        """Called when the application starts."""
        if SpanName.MAIN_FUNCTION in self._spans:
            return  # This callback may be called multiple times

        global _START_TIME
        self._start_span(SpanName.MAIN_FUNCTION, start_time=_START_TIME)

    def on_app_end(self, *args, **kwargs) -> None:  # type: ignore [no-untyped-def]
        """Called when the application ends."""
        if SpanName.MAIN_FUNCTION not in self._spans:
            return  # This callback may be called multiple times

        self._stop_span(SpanName.MAIN_FUNCTION)

    # For checkpoint save callbacks, we add our own wrapper on trainer.save_checkpoint, see setup() and teardown() functions.
    # Nemo callbacks are also available, but it looks like they may be called in child processes and do not pass sufficient
    # information to the callback, so we avoid them. It's important that the callback names do not clash, so we moved the "start"
    # and "end" at the beginning of the function names.
    def on_start_save_checkpoint(self, trainer: Trainer, filepath: _PATH) -> None:
        if self.state.is_async_checkpoint:
            span_name = SpanName.CHECKPOINT_SAVE_ASYNC
        else:
            span_name = SpanName.CHECKPOINT_SAVE_SYNC

        self._start_span(span_name)

    def on_end_save_checkpoint(
        self,
        trainer: Trainer,
        exception: Optional[Exception],
    ) -> None:
        if self.state.is_async_checkpoint:
            span_name = SpanName.CHECKPOINT_SAVE_ASYNC
        else:
            span_name = SpanName.CHECKPOINT_SAVE_SYNC

        metrics = CheckpointMetrics.create(
            current_iteration=trainer.global_step,
        )
        self._stop_span(span_name, metrics)

    # Dataloader lifecycle (from Nemo BaseCallback version >= 25.09) or called manually by the user
    def on_dataloader_init_start(self, *args, **kwargs) -> None:  # type: ignore [no-untyped-def]
        """
        Called when dataloader initialization starts.
        This is not called by PyTorch Lightning, but is called by Nemo version >= 25.09.
        """
        self._start_span(SpanName.DATA_LOADER_INIT)

    def on_dataloader_init_end(self, *args, **kwargs) -> None:  # type: ignore [no-untyped-def]
        """
        Called when dataloader initialization ends.
        This is not called by PyTorch Lightning, but is called by Nemo version >= 25.09.
        """
        self._stop_span(SpanName.DATA_LOADER_INIT)

    # Optimizer lifecycle (from Nemo BaseCallback version >= 25.09) or called manually by the user
    def on_optimizer_init_start(self, *args, **kwargs) -> None:  # type: ignore [no-untyped-def]
        """
        Called when optimizer initialization starts.
        This is not called by PyTorch Lightning, but is called by Nemo version >= 25.09.
        """
        self._start_span(SpanName.OPTIMIZER_INIT)

    def on_optimizer_init_end(self, *args, **kwargs) -> None:  # type: ignore [no-untyped-def]
        """
        Called when optimizer initialization ends.
        This is not called by PyTorch Lightning, but is called by Nemo version >= 25.09.
        """
        self._stop_span(SpanName.OPTIMIZER_INIT)

    # Checkpoint load lifecycle (from Nemo BaseCallback version >= 25.09) or called manually by the user
    def on_load_checkpoint_start(self, *args, **kwargs) -> None:  # type: ignore [no-untyped-def]
        """
        Called when checkpoint loading starts.
        This is not called by PyTorch Lightning, but is called by Nemo version >= 25.09.
        """
        self._start_span(SpanName.CHECKPOINT_LOAD)

    def on_load_checkpoint_end(self, *args, **kwargs) -> None:  # type: ignore [no-untyped-def]
        """
        Called when checkpoint loading ends.
        This is not called by PyTorch Lightning, but is called by Nemo version >= 25.09.
        """
        self._stop_span(SpanName.CHECKPOINT_LOAD)

    # Remaining functions are from PyTorch Lightning
    def on_train_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        self._start_span(SpanName.TRAINING_LOOP)
        self.state.reset()

    def on_train_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        # Log any remaining metrics that haven't been logged yet
        if trainer.global_step != self.state.last_logged_step:
            self._log_iteration_metrics(trainer)
        self._stop_span(SpanName.TRAINING_LOOP)

    def on_train_batch_start(self, trainer: Trainer, pl_module: LightningModule, batch: Any, batch_idx: int) -> None:
        self._start_span(SpanName.ITERATION, verbosity=Verbosity.PROFILING)
        self._start_span(SpanName.MODEL_FORWARD, verbosity=Verbosity.PROFILING)

        if self.batch_flops == 0:
            self._start_monitoring_flops()

    # Note that this won't be called by Nemo when using the Megatron Core models and training strategy. See function docstring for more details.
    def on_before_backward(
        self, trainer: Trainer, pl_module: LightningModule, loss: torch.Tensor, optimizer_idx: int = 0
    ) -> None:
        """
        The before and after backward callbacks are called before and after the backward pass, respectively.
        Note that when using Megatron Core models and Megatron training strategies in Nemo, these will never be called. This is
        because Megatron Core does not invoke model.backward(), it only invokes model.forward() and then uses
        a loss function to compute the backward pass, which cannot be tracked with PTL callbacks. So when
        running Megatron models in Nemo, the forward time actually includes the backward pass. This is the same
        for MLM, where forward-backward are timed together in MLM log messages (when enabled).
        """
        if SpanName.MODEL_FORWARD in self._spans:
            forward_elapsed = self._stop_span(SpanName.MODEL_FORWARD)
            self.state.span_summaries[SpanName.MODEL_FORWARD].add(forward_elapsed)

        span_name = SpanName.MODEL_BACKWARD
        self._start_span(span_name, verbosity=Verbosity.PROFILING)

    def on_after_backward(self, trainer: Trainer, pl_module: LightningModule) -> None:
        span_name = SpanName.MODEL_BACKWARD
        if span_name in self._spans:
            backward_elapsed = self._stop_span(span_name)
            self.state.span_summaries[span_name].add(backward_elapsed)

    def on_before_optimizer_step(
        self, trainer: Trainer, pl_module: LightningModule, optimizer: torch.optim.Optimizer
    ) -> None:
        # Stop MODEL_FORWARD if it's still running (happens with Megatron Core models and training strategies in Nemo that do not invoke model.backward())
        if SpanName.MODEL_FORWARD in self._spans:
            forward_elapsed = self._stop_span(SpanName.MODEL_FORWARD)
            self.state.span_summaries[SpanName.MODEL_FORWARD].add(forward_elapsed)

        if self.batch_flops_monitor is not None:
            self._end_monitoring_flops()

        span_name = SpanName.OPTIMIZER_UPDATE
        self._start_span(span_name, verbosity=Verbosity.PROFILING)

    def on_train_batch_end(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        outputs: STEP_OUTPUT,
        batch: Any,
        batch_idx: int,
    ) -> None:
        # Stop monitoring FLOPS if still running
        if self.batch_flops_monitor is not None:
            self._end_monitoring_flops()

        # Stop MODEL_FORWARD if it's still running (happens with Megatron Core models and training strategies in Nemo that do not invoke model.backward())
        if SpanName.MODEL_FORWARD in self._spans:
            forward_elapsed = self._stop_span(SpanName.MODEL_FORWARD)
            self.state.span_summaries[SpanName.MODEL_FORWARD].add(forward_elapsed)

        # Stop OPTIMIZER_UPDATE if it was started
        if SpanName.OPTIMIZER_UPDATE in self._spans:
            optimizer_update_elapsed = self._stop_span(SpanName.OPTIMIZER_UPDATE)
            self.state.span_summaries[SpanName.OPTIMIZER_UPDATE].add(optimizer_update_elapsed)

        iteration_elapsed = self._stop_span(SpanName.ITERATION)
        self.state.span_summaries[SpanName.ITERATION].add(iteration_elapsed)

        batch_size = extract_batch_size(batch)
        self.state.num_samples += batch_size
        if self.batch_flops > 0:
            self.state.num_flops += self.batch_flops

        # Send an event only when global_step has changed and is a multiple of log_interval.
        # This prevents logging multiple times per optimizer step when using gradient accumulation.
        if trainer.global_step != self.state.last_logged_step and trainer.global_step % self.log_interval == 0:
            self._log_iteration_metrics(trainer)
            self.state.last_logged_step = trainer.global_step

    def _log_iteration_metrics(self, trainer: Trainer) -> None:
        """Log the iteration metrics"""
        iterations = self.state.span_summaries[SpanName.ITERATION].num
        if iterations == 0:
            return

        avg_iteration_time = self.state.span_summaries[SpanName.ITERATION].elapsed / iterations
        avg_forward_time = (
            (
                self.state.span_summaries[SpanName.MODEL_FORWARD].elapsed
                / self.state.span_summaries[SpanName.MODEL_FORWARD].num
            )
            if self.state.span_summaries[SpanName.MODEL_FORWARD].num > 0
            else 0
        )
        avg_backward_time = (
            (
                self.state.span_summaries[SpanName.MODEL_BACKWARD].elapsed
                / self.state.span_summaries[SpanName.MODEL_BACKWARD].num
            )
            if self.state.span_summaries[SpanName.MODEL_BACKWARD].num > 0
            else 0
        )

        avg_optimizer_update_time = (
            (
                self.state.span_summaries[SpanName.OPTIMIZER_UPDATE].elapsed
                / self.state.span_summaries[SpanName.OPTIMIZER_UPDATE].num
            )
            if self.state.span_summaries[SpanName.OPTIMIZER_UPDATE].num > 0
            else 0
        )

        if avg_forward_time > 0 or avg_backward_time > 0:
            # For Megatron models in Nemo, avg_backward_time will be 0 but avg_forward_time will include both
            flops_per_second = self.state.num_flops / (avg_forward_time + avg_backward_time)
            tflops_per_second = flops_per_second / 1e12
        else:
            tflops_per_second = 0

        metrics = IterationMetrics.create(
            current_iteration=trainer.global_step,
            num_iterations=self.log_interval,
            interval=self.log_interval,
            batch_size=int(self.state.num_samples / iterations),
            average_iteration_time=avg_iteration_time,
            tflops=tflops_per_second,
            average_forward_time=avg_forward_time,
            average_optimizer_update_time=avg_optimizer_update_time,
        )

        if avg_backward_time > 0:
            metrics.add_metric("average_backward_time", avg_backward_time)

        self.recorder.event(Event.create(EventName.TRAINING_ITERATIONS, metrics))
        self.state.reset()

    def on_exception(self, trainer: Trainer, pl_module: LightningModule, exception: BaseException) -> None:
        if isinstance(exception, Exception):
            self.recorder.error(str(exception), exception)
        else:
            message = f"{type(exception).__name__}: {str(exception)}\n" + "".join(traceback.format_exception(exception))
            self.recorder.error(message)

    def on_validation_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        self._start_span(SpanName.VALIDATION_LOOP)
        self.state.reset()

    def on_validation_batch_start(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        # Skip because it may not run all callbacks during sanity checking
        if trainer.sanity_checking:
            return

        self._start_span(SpanName.ITERATION, verbosity=Verbosity.PROFILING)

    def on_validation_batch_end(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        outputs: STEP_OUTPUT,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        # Skip because it may not run all callbacks during sanity checking
        if trainer.sanity_checking:
            return

        validation_elapsed = self._stop_span(SpanName.ITERATION)
        self.state.span_summaries[SpanName.ITERATION].add(validation_elapsed)

    def on_validation_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        if self.state.span_summaries[SpanName.ITERATION].num == 0:
            self._stop_span(SpanName.VALIDATION_LOOP)
            return

        avg_validation_time = (
            self.state.span_summaries[SpanName.ITERATION].elapsed / self.state.span_summaries[SpanName.ITERATION].num
        )
        metrics = IterationMetrics.create(
            current_iteration=trainer.global_step,
            num_iterations=self.state.span_summaries[SpanName.ITERATION].num,
            average_iteration_time=avg_validation_time,
        )
        self._stop_span(SpanName.VALIDATION_LOOP, metrics)
        self.state.reset()



#
#  Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an
# express license agreement from NVIDIA CORPORATION is strictly
# prohibited.
#
import importlib
import os
from types import ModuleType
from typing import Any, Optional

__torch_import_attempted: bool = False
__torch_distributed_module: Optional[ModuleType] = None


def import_torch_distributed() -> Optional[ModuleType]:
    """Lazy import of the torch.distributed package to avoid a hard dependency for non-torch applications."""
    global __torch_distributed_module, __torch_import_attempted
    if __torch_import_attempted:
        return __torch_distributed_module

    try:
        __torch_distributed_module = importlib.import_module("torch.distributed")
        return __torch_distributed_module
    except ImportError:
        return None
    finally:
        __torch_import_attempted = True


def get_rank(group: Optional[Any] = None) -> int:
    """Get the rank (GPU device) of the worker.

    Returns:
        rank (int): The rank of the worker.
    """
    rank = int(os.getenv("RANK", "0"))
    dist = import_torch_distributed()
    if dist is not None and dist.is_available() and dist.is_initialized():
        rank = dist.get_rank(group)
    return rank


def get_world_size(group: Optional[Any] = None) -> int:
    """Get world size. How many GPUs are available in this job.

    Returns:
        world_size (int): The total number of GPUs available in this job.
    """
    world_size = int(os.getenv("WORLD_SIZE", "1"))
    dist = import_torch_distributed()
    if dist is not None and dist.is_available() and dist.is_initialized():
        world_size = dist.get_world_size(group)
    return world_size


def is_rank0() -> bool:
    """Check if current process is the master GPU.

    Returns:
        (bool): True if this function is called from the master GPU, else False.
    """
    return get_rank() == 0


def start_monitoring_flops() -> Any:
    """
    If possible, enable pytorch flops counter mode, to track floating point operations that are performed,
    until end_monitoring_tflops is called.
    Returns:
        flop_counter (Any): The flop counter object if successful, else None.
    """
    try:
        from torch.utils.flop_counter import FlopCounterMode

        flop_counter = FlopCounterMode(display=False)
        flop_counter.__enter__()
        return flop_counter
    except Exception:
        return None


def end_monitoring_flops(flop_counter: Any) -> float:
    if flop_counter is not None:
        flop_counter.__exit__(None, None, None)
        return float(flop_counter.get_total_flops())
    else:
        return 0
